%==========================================================================
% Scenario-based video event detection by constraint flow
% Composite event detector
% Instantaneous observations (only primitives as observations)
% 2010. 1. 5. Suha Kwak, POSTECH.
%==========================================================================



function detEvent(event_name)
    % event_name     : the name of the event to be detected

    

%% PRELIMINARY
config
load config

% loading the scenario and primitive detection resutls ====================
evalc(['load .\datasets\scenario_', event_name]);
evalc(['load .\datasets\primitives_', event_name]);

% about probabilities =====================================================
% bound the range of probabilities
prob_eps = 5e-2;

% lower bound of the log posterior probability to avoid machine epsilon problems
log_prob_min = -1e+10;

% penalty weight for staying on idle flow nodes
idle_penalty = 5e-1;

% offline scenario parsing ================================================
disp(['The scenario of the event ', event_name, ' is [', scenario, '].'])
[elem_name, FN_list, FN_link, init_FN_num, dummy_elems] = parser(scenario);
    % elem_name   : a set of names of primitives
    % FN_list     : a list of flow nodes of the scenario
    % FN_link     : a link structure of the flow structure
    % init_FN_num : the number of initial flow nodes = 
    %               the last index of the initial flow nodes in the FN_list
    % dummy_elems : a list of dummy elements (binary)
disp('Scenario parsing is completed.')
    
% the number of primitives
elem_num = length(elem_name);

% the number of flow nodes
FN_num = size(FN_list, 1);
disp([' - the number of primitives : ', num2str(elem_num)])
disp([' - the number of flow nodes : ', num2str(FN_num)])

% FN_list backprojected to binary space
FN_list_bin = (FN_list == cond_A);

% list of idle flow nodes
    % all elements are 'inactive' but not the initial flow node
    % it means that at least one 'finished' element should be on the nodes
FN_list_idle = (sum(FN_list_bin, 2) == 0);
FN_list_idle(1 : init_FN_num) = false;

% the likelihoods of the corresponding primitives =========================
% a sequence of observation probabilities (likelihoods) of the primitives
obs_prob_seq = zeros(frame_num, elem_num);

% select the corresponding primitives
for elem_idx = 1 : elem_num
    if dummy_elems(elem_idx)
        % if the element is a dummy, assign the smallest probability (it does not happen)
        obs_prob_seq(:, elem_idx) = prob_eps;
    else
        % else, assign the corresponding likelihood
        evalc(['obs_prob_seq(:, ', num2str(elem_idx), ') = ', elem_name{elem_idx}]);
    end
end

% data structures to manage samples =======================================
% log posterior probabilties for each sample
DS_posterior = ones(FN_num, 1) .* -inf;

% detection result of the entire event
FN_seq_list = uint16(false(FN_num, frame_num));



%% EVENT DETECTION
% initial probability for the initial flow nodes
DS_posterior(1 : init_FN_num) = log(1 / init_FN_num);
tic
for frm_idx = 1 : frame_num
    disp(['  ', num2str(frm_idx), ' / ', num2str(frame_num), ' frame'])

% online event detection ==================================================
    % current observation probabilities 
    prob_set = repmat(obs_prob_seq(frm_idx, :), FN_num, 1);
    obs_prob = prod((FN_list_bin .* prob_set) + (~FN_list_bin .* (1 - prob_set)), 2);
    
    % buffers for prediction (before filtering)
    DS_posterior_buf = ones(FN_num, 1) .* -inf;
    FN_seq_list_buf = zeros(FN_num, frame_num, 'uint16');

    for fn_idx = 1 : FN_num
        % find the predecessor node whose posterior is maximal
        pre_FN_prob = DS_posterior;
        pre_FN_prob(~FN_link(:, fn_idx)) = -inf;
        [max_pre_FN_prob, max_pre_FN_idx] = max(pre_FN_prob);
        
        % interpretation update and search for the best prediction
        if isinf(max_pre_FN_prob)
            continue;
        end
        DS_posterior_buf(fn_idx) = max_pre_FN_prob;
        FN_seq_list_buf(fn_idx, :) = FN_seq_list(max_pre_FN_idx, :);
        FN_seq_list_buf(fn_idx, frm_idx) = fn_idx;
    end

    % confirm the information about temporal intervals
    FN_seq_list = FN_seq_list_buf;
    
    % validation by idle penalty and observation probabilities
    DS_posterior_buf = DS_posterior_buf - (FN_list_idle .* idle_penalty);
    DS_posterior = DS_posterior_buf + log(obs_prob);
    
    % to avoid machine epsilon problem
    min_post = min(DS_posterior);
    if min_post < log_prob_min && min_post ~= -inf
        DS_posterior = DS_posterior - log_prob_min;
    end
end
toc



%% DETECTION RESULTS
% generating the most probable interpretation from "best_FN_seq"
[~, best_FN_idx] = max(DS_posterior);
best_FN_seq = FN_seq_list(best_FN_idx, :);
best_intp = FN_list_bin(best_FN_seq, :);

% plot the estimated intervals of the primitives 
h = figure;
set(h, 'Position', [300, 400, 1500, 510])
hold on;
for elem_idx = 1 : elem_num
    bar_height = elem_idx / (elem_num + 1);
    elem_intv = find(best_intp(1 : frame_num, elem_idx));
    plot(elem_intv, ones(length(elem_intv), 1) .* bar_height, 'sk', 'MarkerEdgeColor', 'k', 'MarkerFaceColor', [0 0 0], 'MarkerSize', 5);
end
hold off;
set(gca,'ytick', 0 : 1 / (elem_num + 1) : 1)
ytickvals = 0 : 1 / (elem_num + 1) : 1;
yticklabels = cell(length(ytickvals), 1);
yticklabels{1} = '';
for idx = 1 : elem_num
    yticklabels{idx + 1} = elem_name{idx};
end
yticklabels{end} = '';
set(gca,'YTickLabel', yticklabels)
xlabel('Frames')
title(['Detection result of ', event_name])

