%==========================================================================
% Semantic Flow for Complex-Event Recognition in Videos
% Generate flow nodes by all possible combination of shifted conditions
% 2009. 12. 15. Suha Kwak, POSTECH.
%==========================================================================



function gen_FN = genFNodes(cur_FN, cond_shift_map, iterative_elems, lambda_group_list, init_FN)
    % cur_FN            : the current flow node, a vector
    % cond_shift_map    : a set of possible condition lists for the elements
    % iterative_elems   : a list of elements which can be occurred iteratively, grouped by ranges of '*' operators
    % lambda_group_list : a list of groups on which the lambda-transition can be applied
    % init_FN           : the initial flow nodes
    


%% PRELIMINARY
% load configuration
load config

% the number of elements
elem_num = length(cur_FN);

% the number of groups
group_num = size(iterative_elems, 1);

% initialization of the list of generated flow nodes 
% the parser should guarantee self-transitions of flow nodes always
gen_FN = cur_FN;
gen_FN_num = 1;



%% FLOW NODE GENERATION BY CONDITION SHIFTS
% generate all possible combinations of condition shifts (for each element)
for elem_idx = 1 : elem_num
    cand_cond = find(cond_shift_map(elem_idx, :));

    gen_FN_num_temp = gen_FN_num;
    for fn_idx = 1 : gen_FN_num
        prev_FN = gen_FN(fn_idx, :);
        for cond_idx = 1 : length(cand_cond)
            new_FN = prev_FN;
            new_FN(elem_idx) = cand_cond(cond_idx);
            gen_FN = [gen_FN ; new_FN];
            gen_FN_num_temp = gen_FN_num_temp + 1;
        end
    end
    gen_FN_num = gen_FN_num_temp;
end



%% FLOW NODE GENERATION BY LAMBDA-TRANSITIONS
% generate all possible combinations of group condition shifts (for each star group)
for group_idx = 1 : group_num
    if ~lambda_group_list(group_idx)
        continue;
    end
    
    own_cond_I = cond_I + group_idx;
    gen_FN_num_temp = gen_FN_num;
    for fn_idx = 1 : gen_FN_num
        new_FN = gen_FN(fn_idx, :);
        new_FN(iterative_elems(group_idx, :)) = own_cond_I;
        gen_FN = [gen_FN ; new_FN];
        gen_FN_num_temp = gen_FN_num_temp + 1;
    end
    gen_FN_num = gen_FN_num_temp;
end



%% CHECK DUMMY ELEMENTS, LAST AND DUPLICATED FLOW NODES
% dummy setting
gen_FN(:, (cur_FN == cond_E)) = cond_E;

% dulplicated flow nodes --------------------------------------------------
% the number of generated flow nodes
gen_FN_num = size(gen_FN, 1);

% a list of valid flow nodes
valid_FN_list = true(gen_FN_num, 1);

% remove duplicated flow nodes
for gen_fn_idx1 = 1 : gen_FN_num - 1
    val_FN = gen_FN(gen_fn_idx1, :);
    for gen_fn_idx2 = gen_fn_idx1 + 1 : gen_FN_num
        if sum(val_FN == gen_FN(gen_fn_idx2, :)) == elem_num
            valid_FN_list(gen_fn_idx2) = false;
        end
    end
end

% remove invalid flow nodes
gen_FN = gen_FN(valid_FN_list, :);



%% FLOW NODE GENERATION BY RECURSIVE LAMBDA-TRANSITIONS
% whether the last nodes is in the generated flow nodes or not
last_FN_is = false;

% the number of generated flow nodes
gen_FN_num = size(gen_FN, 1);

% a list of valid flow nodes
valid_FN_list = true(gen_FN_num, 1);

for gen_fn_idx = 1 : gen_FN_num
    val_FN = gen_FN(gen_fn_idx, :);
    % val_FN is one of the last nodes
    if sum(val_FN == cond_F | val_FN == cond_E) == elem_num
        last_FN_is = true;
        valid_FN_list(gen_fn_idx) = false;
    end    
end

% lambda transition from the last flow nodes to the initial flow nodes
gen_FN = gen_FN(valid_FN_list, :);
if last_FN_is
    gen_FN = [gen_FN ; init_FN];
end


