%==========================================================================
% Scenario-Based Video Event Recognition by Constraint Flow
% Main scenario parser
% 2009. 12.  3. Suha Kwak, POSTECH.
% 2013.  6. 16. Updated by Suha Kwak
%==========================================================================



function [elem_name, FN_list, FN_link, init_FN_num, dummy_elems, anc_list] = parser(strin)
    % strin       : input string represent an event scenario
    % elem_name   : a set of names of element events
    % FN_list     : a list of flow nodes of the scenario
    % FN_link     : a link structure of the flow structure
    % init_FN_num : the number of initial flow nodes
    %               = the last index of the initial flow nodes in the FN_list
    % dummy_elems : a list of dummy elements (binary)
    % anc_list    : a set of ancestor lists for the elements





%% PRELIMINARY
    % 1. elementary event identification and segmentation    
    % 2. simplified form of the input scenario (symbolic string)
    
% load configuration
load config
    
% determine primitive events ----------------------------------------------
elem_name_comp = bwconncomp(isletter(strin) | strin == dummy);
elem_num = length(elem_name_comp.PixelIdxList);
elem_name = cell(elem_num, 1);

for eidx = 1 : elem_num
    elem_name{eidx} = strin(elem_name_comp.PixelIdxList{eidx});
end

% dummy identification: list of dummy primitives
dummy_elems = cellfun(@(x) x(1) == dummy, elem_name);


% symbolic string for easier parsing --------------------------------------
symstr = strin;
for eidx = 1 : elem_num
    ename = elem_name{eidx};
    elist = strfind(symstr, ename);
    symstr = [symstr(1 : elist(1) - 1), ...
              num2str(eidx, '%03d'), ...
              symstr(elist(1) + length(ename) : end)];
end





%% OR GROUPING TO GENERATE MULTIPLE FLOW STRUCTURES
    % generate multiple initial flow nodes > multiple flow structures

% the number of OR groups
or_group_num = 0;

% structure for OR grouping results
or_group_map = uint16(false(0));

% find '|' operators
oper_pos = (symstr == selection);

% iterative grouping
while sum(oper_pos) > 0
    cur_oper_pos = find(oper_pos, 1, 'first');
    oper_pos(cur_oper_pos) = false;
    or_group_num = or_group_num + 1;
        
% fore part of the current '|' operator ===================================
    % the number of symbolic parts in this OR group
    part_num = 1;
    fore_part = foreEffPart(symstr(1 : cur_oper_pos - 1));
    
    sym_comp = bwconncomp(fore_part > 47 & fore_part < 58);
    nsym = length(sym_comp.PixelIdxList);
    for sidx = 1 : nsym
        elem_idx = str2double(fore_part(sym_comp.PixelIdxList{sidx}));
        or_group_map(or_group_num, elem_idx) = part_num;
    end
    
% post traversing =========================================================
    part_num = 2;
    head_pos = cur_oper_pos + 1;
    post_part = hindEffPart(symstr(head_pos : length(symstr)));
    paren_num = 0;
    pidx = 1;
    while pidx <= length(post_part)        
% part included in a pair of parentheses ----------------------------------
        if post_part(pidx) == '('
            paren_num = paren_num + 1;
        elseif post_part(pidx) == ')'
            paren_num = paren_num - 1;
            if paren_num < 0
                disp('parser: wrong input string (# of parentheses)');
                return;
            end
            
% another '|' operator - nested or not? -----------------------------------
        elseif post_part(pidx) == selection
            % nested case : just skip
            if paren_num > 0
            % not nested : should find the next part
            else
                oper_pos(cur_oper_pos + pidx) = false;
                part_num = part_num + 1;
            end
            
% primitives - update -----------------------------------------------------
        elseif post_part(pidx) > 47 && post_part(pidx) < 58
            % find all digits of the primitive
            qidx = pidx + 1;
            while (post_part(qidx + 1) > 47)  && (post_part(qidx + 1) < 58)
                qidx = qidx + 1;
                if qidx == length(post_part)
                    break;
                end
            end
            % save
            elem_idx = str2double(post_part(pidx : qidx));
            or_group_map(or_group_num, elem_idx) = part_num;            
            pidx = qidx;            
            
% other operators are not important in this case --------------------------
        else
            % skip
        end
        
        pidx = pidx + 1;
    end
end

% size calibration of the OR grouping map (do we need this?)
if or_group_num > 0 && size(or_group_map, 2) < elem_num
    or_group_map(or_group_num, elem_num) = 0;
end





%% ELEMENT GROUPING TO HANDLE COMPLEX * STRUCTURES
    % for the unary relation operator '*'

% position of the '*' operators
oper_pos = find(symstr == iter_occur);

% number of * groups
iterative_group_num = length(oper_pos);

% the list of elements which can be occurred iteratively
% -- one element can be included by multiple '*' groups
iterative_elems = false(iterative_group_num, elem_num);

% the list of temporally last elements of each group of '*' range 
    % (a group of elements which can be occurred iteratively by the same '*')
iterative_elems_last = false(iterative_group_num, elem_num);


% grouping by each '*' operator ===========================================
for oper_idx = 1 : iterative_group_num
    str_idx = oper_pos(oper_idx) - 1;

% '*' group: only an element ----------------------------------------------
    if symstr(str_idx) > 47 && symstr(str_idx) < 58
        % find all digits of the primitive
        hidx = str_idx - 1;
        while symstr(hidx - 1) > 47 && symstr(hidx - 1) < 58
            hidx = hidx - 1;
            if hdix == 1
                break;
            end
        end
        % save
        star_elem_idx = str2double(symstr(hidx : str_idx));
        iterative_elems(oper_idx, star_elem_idx) = true;
        iterative_elems_last(oper_idx, star_elem_idx) = true;

% '*' group: elements bounded by a pair of parentheses --------------------
% -- an effect range of an operator '*' is bounded only by a pair of parentheses
    elseif symstr(str_idx) == ')'
        % backtracking until getting the corresponding parenthesis
        hidx = str_idx;
        paren_num = 1;
        while paren_num > 0
            hidx = hidx - 1;
            if symstr(hidx) == '('
                paren_num = paren_num - 1;
            elseif symstr(hidx) == ')'
                paren_num = paren_num + 1;
            end
        end
        star_part = symstr(hidx : str_idx);
        % find elements in the range
        star_elem_comp = bwconncomp((star_part > 47) & (star_part < 58));
        star_elems = cellfun(@(x) str2double(star_part(x)), star_elem_comp.PixelIdxList);
        iterative_elems(oper_idx, star_elems) = true;        
        % find the last elements in the range
        star_elems_last = lastPrim(star_part);
        iterative_elems_last(oper_idx, star_elems_last) = true;        
        
% invalid input string ----------------------------------------------------
    else
        disp('parser: wrong input string (near a unary relation)');
        return
    end
end





%% FIND STARTING TIME ANCESTORS AND ENDING TIME ANCESTORS
    % find temporal ancestors restricting a starting and an ending time of each element     

% a set of ancestor lists for each element
anc_list = uint8(zeros(elem_num));
    % row    : each element
    % column : ancestors
    % value  : relationships (0 : no anc, 1~ : relations with the anc's)

% binary relation operator '<' ============================================
oper_pos = find(symstr == temp_nonadj);
for oper_idx = 1 : length(oper_pos)
    % fore and hind effective parts of the operator
    fore_effpart = foreEffPart(symstr(1 : (oper_pos(oper_idx) - 1)));
    hind_effpart = hindEffPart(symstr((oper_pos(oper_idx) + 1) : length(symstr)));
    
    % '<' connects ending times of fore elements with starting times of hind elements
    fore_last_elems = lastPrim(fore_effpart);
    hind_first_elems = firstPrim(hind_effpart);
    
    % update ancestors
    anc_list(hind_first_elems, fore_last_elems) = stt_TN;
end


% binary relation operator '^' ============================================
oper_pos = find(symstr == during);
for oper_idx = 1 : length(oper_pos)
    % fore and hind effective parts of the operator
    fore_effpart = foreEffPart(symstr(1 : (oper_pos(oper_idx) - 1)));
    hind_effpart = hindEffPart(symstr((oper_pos(oper_idx) + 1) : length(symstr)));
    
% order of starting times -------------------------------------------------    
    % '^' connects starting times of fore elements with starting times of hind elements
    fore_first_elems = firstPrim(fore_effpart);
    hind_first_elems = firstPrim(hind_effpart);
    
    % update starting time ancestors
    anc_list(hind_first_elems, fore_first_elems) = stt_DU;
    
% order of ending times ---------------------------------------------------
    % '^' connects ending times of hind elements with ending times of fore elements
    fore_last_elems = lastPrim(fore_effpart);
    hind_last_elems = lastPrim(hind_effpart);
    
    % update ending time ancestors
    anc_list(fore_last_elems, hind_last_elems) = end_DU;
end


% binary relation operator '~' ============================================
oper_pos = find(symstr == start_ahead);
for oper_idx = 1 : length(oper_pos)
    % fore and hind effective parts of the operator
    fore_effpart = foreEffPart(symstr(1 : (oper_pos(oper_idx) - 1)));
    hind_effpart = hindEffPart(symstr((oper_pos(oper_idx) + 1) : length(symstr)));
    
    % '~' connects starting times of fore elements with starting times of hind elements
    fore_first_elems = firstPrim(fore_effpart);
    hind_first_elems = firstPrim(hind_effpart);
    
    % update starting time ancestors
    anc_list(hind_first_elems, fore_first_elems) = stt_SA;
end


% handling dummy ancestors ================================================
% if an ancestor of an element is a dummy, 
% then the element should serve the dummy's ancestors as its ancestors.
dummy_list = find(dummy_elems);
for dm_idx = 1 : length(dummy_list)
    elems_anc_dummy = find(anc_list(:, dummy_list(dm_idx)) > 0);
    anc_list(elems_anc_dummy, :) = anc_list(elems_anc_dummy, :) + ...
                                   repmat(anc_list(dummy_list(dm_idx), :), length(elems_anc_dummy), 1);
end






%% PARSING - TRAVERSING THE CONSTRAINT FLOW
% Similar with BFS(breadth-first search) algorithm
    % 1. for each flow node currently maintained, check all possible condition-shifts of each element
        % 1-1. possible condition shift: {cond_R->cond_A, cond_A->cond_F, cond_I->cond_A, lambda-transitions}
        % 1-2. self-shift should be possible naturally
    % 2. generate successor nodes by using all possible combinations of checked conditions for each element
    % 3. update the link structure: {a link between current node and a successor node}
    

% helpers -----------------------------------------------------------------
% a list of possible flow nodes
FN_list = uint16(zeros(0));
FN_num = 0;

% link information (where a flow node can GO TO)
FN_link = false(0);

% flow nodes the parser visits currently
FN_queue = uint16(zeros(0));
queue_num = 0;


% initialization ----------------------------------------------------------

% initial flow node generation
init_FN = genInitFNodes(dummy_elems, or_group_map);
init_FN_num = size(init_FN, 1);

% initialization of the list and the link information
FN_num = FN_num + init_FN_num;
FN_list(1 : init_FN_num, :) = init_FN;

% push the initial flow node to the queue
queue_num = queue_num + init_FN_num;
for fn_idx = 1 : init_FN_num
    FN_queue(fn_idx, :) = [fn_idx, init_FN(fn_idx, :)];  % -- [index, flow node]
end


while queue_num > 0
    % pop from the queue
    pop_info = FN_queue(1, :);
    FN_queue = FN_queue(2 : queue_num, :);
    queue_num = queue_num - 1;
    cur_FN_idx = pop_info(1);
    cur_FN = pop_info(2 : length(pop_info));    

    
% checking condition shifts >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    % a set of possible condition lists for the elements (except self-transition)
    cond_shift_map = false(elem_num, cond_num);
    
    % validate condition shifts
    for elem_idx = 1 : elem_num
        % are condition-shifts possible?
        ignition = true;
        
        % temporal ancestors of the element
        stt_anc_TN = find(anc_list(elem_idx, :) == stt_TN);
        stt_anc_DU = find(anc_list(elem_idx, :) == stt_DU);
        stt_anc_SA = find(anc_list(elem_idx, :) == stt_SA);
        end_anc_DU = find(anc_list(elem_idx, :) == end_DU);
        
% cond_R: can be shifted to cond_A ========================================
        if cur_FN(elem_idx) == cond_R
            % '*' group indices
            iter_group_idx = find(iterative_elems(:, elem_idx));
            
% starting time restriction by '<' ----------------------------------------
            for anc_idx = 1 : length(stt_anc_TN) 
                if cur_FN(stt_anc_TN(anc_idx)) == cond_R || cur_FN(stt_anc_TN(anc_idx)) == cond_A
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_TN(anc_idx)) > cond_I
                    anc_iter_group_idx = cur_FN(stt_anc_TN(anc_idx)) - cond_I;
                    % the element should be outside the '*' range of the ancestor
                    % when it is on 'cond_R' and the ancestor is on 'cond_I' 
                    % because of the "lambda-transition" (all 'cond_I')
                        % therefore, this 'cond_I' is same with 'cond_A'
                    if sum(iter_group_idx == anc_iter_group_idx) == 0
                        ignition = false;
                        break;
                    % else, it is something wrong
                    else
                        disp('parser: parsing error (may be caused by the lambda-transition routine)');
                        disp('      : current element = cond_R, stt_anc_TN = cond_I');
                        return
                    end
                end
            end
            
% starting time restriction by '^' ----------------------------------------
            for anc_idx = 1 : length(stt_anc_DU)
                if cur_FN(stt_anc_DU(anc_idx)) == cond_R
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_DU(anc_idx)) > cond_I
                    anc_iter_group_idx = cur_FN(stt_anc_DU(anc_idx)) - cond_I;
                    % the element should be outside the '*' range of the ancestor
                    % when it is on 'cond_R' and the ancestor is on 'cond_I'
                    % because of the "lambda-transition" (all 'cond_I')
                        % therefore, this 'cond_I' is same with 'cond_A'
                        % >>> it's O.K. (the ignition is possible)
                    % otherwise (the element is inside the '*' range of the ancestor)
                        % it is something wrong
                    if sum(iter_group_idx == anc_iter_group_idx) > 0
                        disp('parser: parsing error (may be caused by the lambda-transition routine)');
                        disp('      : current element = cond_R, stt_anc_DU = cond_I');
                        return
                    end
                end
            end
            
% starting time restriction by '~' ----------------------------------------
            for anc_idx = 1 : length(stt_anc_SA) 
                if cur_FN(stt_anc_SA(anc_idx)) == cond_R
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_SA(anc_idx)) > cond_I
                    anc_iter_group_idx = cur_FN(stt_anc_SA(anc_idx)) - cond_I;
                    % the element should be outside the '*' range of the ancestor
                    % when it is on 'cond_R' and the ancestor is on 'cond_I'
                    % because of the "lambda-transition" (all 'cond_I')
                        % therefore, this 'cond_I' is same with 'cond_A'
                        % >>> it's O.K. (the ignition is possible)
                    % otherwise (the element is inside the '*' range of the ancestor)
                        % it is something wrong
                    if sum(iter_group_idx == anc_iter_group_idx) > 0
                        disp('parser: parsing error (may be caused by the lambda-transition routine)');
                        disp('      : current element = cond_R, stt_anc_SA = cond_I');
                        return
                    end
                end
            end
            
% if there is no restriction anymore, shift the condition -----------------
            if ignition
                cond_shift_map(elem_idx, cond_A) = true;
            end
            
            
% cond_A: can be shifted to cond_F ========================================
        elseif cur_FN(elem_idx) == cond_A
            
% ending time restriction by '^' ------------------------------------------
            for anc_idx = 1 : length(end_anc_DU) 
                if cur_FN(end_anc_DU(anc_idx)) == cond_R || cur_FN(end_anc_DU(anc_idx)) == cond_A
                    ignition = false;
                    break;
                elseif cur_FN(end_anc_DU(anc_idx)) > cond_I
                    % when an element is on 'cond_A', its ending time ancestor can be on any 'cond_I's 
                    % (both of the same range and a difference range)
                    % however, a 'cond_I' is represented as 'cond_R' or 'cond_A' by the situation 
                    % and both of them cannot release ending time of the element on 'cond_A'
                    ignition = false;
                    break;
                end
            end
            
% if there is no restriction anymore, shift the condition -----------------
            if ignition
                cond_shift_map(elem_idx, cond_F) = true;
            end

            
% cond_I: can be shifted to cond_A ========================================
        elseif cur_FN(elem_idx) > cond_I
            % '*' group index for the current 'cond_I'
            iter_group_idx = cur_FN(elem_idx) - cond_I;
            
% re-starting time restriction by '<' -------------------------------------
            for anc_idx = 1 : length(stt_anc_TN) 
                if cur_FN(stt_anc_TN(anc_idx)) == cond_R || cur_FN(stt_anc_TN(anc_idx)) == cond_A
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_TN(anc_idx)) > cond_I                    
                    % when an element is on 'cond_I', its ending time ancestor can be on any 'cond_I's 
                    % (both of the same range and a difference range)
                    % however, a 'cond_I' is represented as 'cond_R' or 'cond_A' by the situation 
                    % and both of them cannot release starting time of the element on 'cond_I'
                    ignition = false;
                    break;
                end
            end
            
% re-starting time restriction by '^' -------------------------------------
            for anc_idx = 1 : length(stt_anc_DU)
                if cur_FN(stt_anc_DU(anc_idx)) == cond_R
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_DU(anc_idx)) > cond_I
                    anc_iter_group_idx = cur_FN(stt_anc_DU(anc_idx)) - cond_I;
                    % same group: the ancestor's 'cond_I' is same with 'cond_R'
                    if iter_group_idx == anc_iter_group_idx
                        ignition = false;
                        break;
                    % differenct group: the ancestor's 'cond_I' is same with 'cond_A'
                    else
                        % it's O.K. (the ignition is possible)
                    end
                end
            end
            
% re-starting time restriction by '~' -------------------------------------
            for anc_idx = 1 : length(stt_anc_SA) 
                if cur_FN(stt_anc_SA(anc_idx)) == cond_R
                    ignition = false;
                    break;
                elseif cur_FN(stt_anc_SA(anc_idx)) > cond_I
                    anc_iter_group_idx = cur_FN(stt_anc_SA(anc_idx)) - cond_I;
                    % same group: the ancestor's 'cond_I' is same with 'cond_R'
                    if iter_group_idx == anc_iter_group_idx
                        ignition = false;
                        break;
                    % different group: the ancestor's 'cond_I' is same with 'cond_A'
                    else
                        % it's O.K. (the ignition is possible)
                    end
                end
            end
            
% if there is no restriction anymore, shift the condition -----------------
            if ignition
                cond_shift_map(elem_idx, cond_A) = true;
            end


% finish and excluded conditions : cannot be shifted ======================
        else
            continue;
        end
    end

    
% check the lambda-transitions ============================================
    % a list of groups on which the lambda-transition can be applied
    lambda_group_list = false(1, iterative_group_num);
    
    % possibility of the lambda-transition for each group
    for group_idx = 1 : iterative_group_num
        gelem_num = sum(iterative_elems(group_idx, :));
        group_cond = cur_FN(iterative_elems(group_idx, :));
        group_cond_last = cur_FN(iterative_elems_last(group_idx, :));
        % to apply the lambda-transition
            % 1. non-last elements of the group are on 'cond_F' or 'cond_E'
            % 2. the last elements of the group are on 'cond_A' or 'cond_F' or 'cond_E'
                % 2-1. at least one last element should be on 'cond_A'        
        if sum(group_cond_last == cond_A) == sum(group_cond == cond_A)                                    && ... % condition 1.
           sum(group_cond == cond_A) + sum(group_cond == cond_F) + sum(group_cond == cond_E) == gelem_num && ... % condition 1, 2.
           sum(group_cond_last == cond_A) > 0                                                                    % condition 2-1.
            lambda_group_list(group_idx) = true;
        end
    end
    


% flow node generation and link structure update >>>>>>>>>>>>>>>>>>>>>>>>>>

% node generation =========================================================
    % generation by all possible combinations of shifts
    gen_FN = genFNodes(cur_FN, cond_shift_map, iterative_elems, lambda_group_list, init_FN);
    gen_FN_num = size(gen_FN, 1);
    

% validation of successors ================================================
    % a list of valid flow nodes
    valid_FN_list = true(gen_FN_num, 1);

    % remove duplicated flow nodes (if they exist)
        % in fact, if the preprocess works correctly, there should not be any duplicated flow nodes
    for gen_fn_idx1 = 1 : gen_FN_num - 1
        val_FN = gen_FN(gen_fn_idx1, :);
        for gen_fn_idx2 = gen_fn_idx1 + 1 : gen_FN_num
            if sum(val_FN == gen_FN(gen_fn_idx2, :)) == elem_num
                disp('parser: warning (duplicated flow nodes)');
                valid_FN_list(gen_fn_idx2) = false;
            end
        end
    end
    
    % remove invalid flow nodes
    gen_FN = gen_FN(valid_FN_list, :);
    gen_FN_num = size(gen_FN, 1);
    
    
% update the link structure and the queue =================================
    for gen_fn_idx = 1 : gen_FN_num
        suc_FN = gen_FN(gen_fn_idx, :);
        
        % check whether suc_FN already exist in the list or not
        exist_FN = false;
        for fn_idx = 1 : FN_num
            if sum(suc_FN == FN_list(fn_idx, :)) == elem_num
                exist_FN = true;
                suc_FN_idx = fn_idx;
                break;
            end
        end
        if exist_FN
            % update link information only
            FN_link(cur_FN_idx, suc_FN_idx) = true;
        else
            % push the successor node to the list of flow nodes
            FN_num = FN_num + 1;
            suc_FN_idx = FN_num;
            FN_list(suc_FN_idx, :) = suc_FN;
            % update link information
            FN_link(cur_FN_idx, suc_FN_idx) = true;
            % push the successor node to the queue
            queue_num = queue_num + 1;
            FN_queue(queue_num, :) = [suc_FN_idx, suc_FN];
        end
    end
end





